###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for:
LHCbDIRAC.HCbDIRAC.Core.Utilities.GeneratorLog"""
import pytest
from LHCbDIRAC.Core.Utilities.GeneratorLog import (
    counterJson,
    efficiencyJson,
    fractionJson,
    crossSectionJson,
    methodGeneratorJson,
)  # noqa


listCounters = [
    {"@name": "all events (including empty events)", "value": "556"},
    {"@name": "events with 0 interaction", "value": "51"},
    {"@name": "generated events", "value": "505"},
    {"@name": "generated interactions", "value": "1395"},
    {"@name": "generated interactions with >= 1b", "value": "7"},
    {"@name": "generated interactions with >= 3b", "value": "1"},
    {"@name": "generated interactions with 1 prompt B", "value": "0"},
    {"@name": "generated interactions with >= 1c", "value": "91"},
    {"@name": "generated interactions with >= 3c", "value": "10"},
    {"@name": "generated interactions with >= prompt C", "value": "2"},
    {"@name": "generated interactions with b and c", "value": "1"},
    {"@name": "accepted events", "value": "1"},
]

dictCounters = {
    "accepted events": 1,
    "all events (including empty events)": 556,
    "events with 0 interaction": 51,
    "generated events": 505,
    "generated interactions": 1395,
    "generated interactions with 1 prompt B": 0,
    "generated interactions with >= 1b": 7,
    "generated interactions with >= 1c": 91,
    "generated interactions with >= 3b": 1,
    "generated interactions with >= 3c": 10,
    "generated interactions with >= prompt C": 2,
    "generated interactions with b and c": 1,
}


@pytest.mark.parametrize("test_input,expected", [(listCounters, dictCounters), ([], {})])
def test_counterJson(test_input, expected):
    assert counterJson(test_input) == expected


listEfficiencies = [
    {"@name": "generator level cut", "after": "1", "before": "4", "error": "0.21651", "value": "0.25"},
    {"@name": "generator particle level cut", "after": "0", "before": "1", "error": "0", "value": "0"},
    {"@name": "generator anti-particle level cut", "after": "1", "before": "3", "error": "0.27217", "value": "0.33333"},
]

dictEfficiencies = {
    "generator anti-particle level cut": {"after": 1, "before": 3, "error": 0.27217, "value": 0.33333},
    "generator level cut": {"after": 1, "before": 4, "error": 0.21651, "value": 0.25},
    "generator particle level cut": {"after": 0, "before": 1, "error": 0.0, "value": 0.0},
}


@pytest.mark.parametrize("test_input,expected", [(listEfficiencies, dictEfficiencies), ([], {})])
def test_efficiencyJson(test_input, expected):
    assert efficiencyJson(test_input) == expected


listFractions = [
    {"@name": "accepted B+", "error": "0.35355", "number": "1", "value": "0.5"},
    {"@name": "accepted B-", "error": "0", "number": "1", "value": "1"},
    {"@name": "accepted B0", "error": "0.35355", "number": "1", "value": "0.5"},
]

dictFractions = {
    "accepted B+": {"error": 0.35355, "number": 1, "value": 0.5},
    "accepted B-": {"error": 0.0, "number": 1, "value": 1.0},
    "accepted B0": {"error": 0.35355, "number": 1, "value": 0.5},
}


@pytest.mark.parametrize("test_input,expected", [(listFractions, dictFractions), ([], {})])
def test_fractionJson(test_input, expected):
    assert fractionJson(test_input) == expected


listCrossSections = [
    {"@id": "101", "description": '"non-diffractive"', "generated": "776", "value": "52.116"},
    {"@id": "102", "description": '"A B -> A B elastic"', "generated": "300", "value": "19.892"},
    {"@id": "103", "description": '"A B -> X B single diffractive"', "generated": "120", "value": "6.24"},
]

dictCrossSections = {
    "A B -> A B elastic": {"ID": 102, "generated": 300, "value": 19.892},
    "A B -> X B single diffractive": {"ID": 103, "generated": 120, "value": 6.24},
    "non-diffractive": {"ID": 101, "generated": 776, "value": 52.116},
}


@pytest.mark.parametrize("test_input,expected", [(listCrossSections, dictCrossSections), ([], {})])
def test_crossSectionJson(test_input, expected):
    assert crossSectionJson(test_input) == expected


listMethods = ["GenerationSignal.SignalPlain", "Generation.SignalRepeatedHadronization"]
listGenerators = ["Redecay", "Pythia8"]
dictMethodsGenerators = {"GenerationSignal.SignalPlain": "Redecay", "Generation.SignalRepeatedHadronization": "Pythia8"}

Method = "Generation.SignalRepeatedHadronization"
Generator = "Pythia8"
dictMethodGenerator = {"Generation.SignalRepeatedHadronization": "Pythia8"}


@pytest.mark.parametrize(
    "test_input1,test_input2,test_input3,expected",
    [(listMethods, listGenerators, 2, dictMethodsGenerators), (Method, Generator, 1, dictMethodGenerator)],  # noqa
)
def test_methodGeneratorJson(test_input1, test_input2, test_input3, expected):
    assert methodGeneratorJson(test_input1, test_input2, test_input3) == expected



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test of the ProductionRequest and Production modules."""
import pytest

from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production

# Production.py
prod = Production()


@pytest.mark.parametrize(
    "input, expected",
    [
        (["T1", "T2"], [{"outputDataType": "t1"}, {"outputDataType": "t2"}]),
        (["T1", "HIST"], [{"outputDataType": "t1"}, {"outputDataType": "hist"}]),
    ],
)
def test__constructOutputFilesList(input, expected):
    res = prod._constructOutputFilesList(input)
    assert res == expected



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Create a new dataset replication or removal transformation according to
plugin."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.TransformationSystem.Utilities.PluginScript import PluginScript

    pluginScript = PluginScript()
    pluginScript.registerPluginSwitches()
    pluginScript.registerFileSwitches()
    Script.registerSwitch("", "Name=", "   Give a name to the transformation, only if files are given")
    Script.registerSwitch(
        "",
        "SetInvisible",
        "Before creating the transformation, set the files in the BKQuery as invisible (default for DeleteDataset)",
    )
    Script.registerSwitch("S", "Start", "   If set, the transformation is set Active and Automatic [False]")
    Script.registerSwitch("", "Force", "   Force transformation to be submitted even if no files found")
    Script.registerSwitch("", "Test", "   Just print out but not submit")
    Script.registerSwitch("", "NoFCCheck", "   Suppress the check in FC for removal transformations")
    Script.registerSwitch("", "Unique", "   Refuses to create a transformation with an existing name")
    Script.registerSwitch("", "Depth=", "   Depth in path for replacing /... in processing pass")
    Script.registerSwitch("", "Chown=", "   Give user/group for chown of the directories of files in the FC")
    Script.registerSwitch(
        "", "MCVersion=", "   (list of) BK ConfigVersion; gets active MC processing passes ('All' for all years)"
    )
    Script.registerSwitch("", "CheckMCReplication", "   List all MC replication transformations that are obsolete")
    Script.registerSwitch("", "ListProcessingPasses", "   Only lists the processing passes")

    Script.registerSwitch(
        "",
        "BodyPlugin=",
        "   BodyPlugin to use for the transformation",
    )
    Script.registerSwitch("", "TransBody=", "   Body to use for the transformation")

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.DataManagementSystem.Client.AddTransformation import executeAddTransformation

    executeAddTransformation(pluginScript)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Agent.ShiftDBAgent.

ShiftDBAgent.__bases__:
  DIRAC.Core.Base.AgentModule.AgentModule
"""
# FIXME: should add a "DryRun" option to run in certification setup
from urllib.request import urlopen
from urllib.error import URLError
import json
import suds.client

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

AGENT_NAME = "ResourceStatus/ShiftDBAgent"


class ShiftDBAgent(AgentModule):
    """This agent queries the LHCb ShiftDB and gets the emails of each piquet
    Then, populates the eGroup associated
    The e-groups admin should be : lhcb-grid-experiment-egroup-admins
    """

    def __init__(self, *args, **kwargs):
        AgentModule.__init__(self, *args, **kwargs)

        # Members initialization

        # ShiftDB url where to find shifter emails
        self.lbshiftdburl = "https://lbshiftdb.cern.ch/list_email"
        # soap wsdl to access eGroups
        self.wsdl = "https://foundservices.cern.ch/ws/egroups/v1/EgroupsWebService/EgroupsWebService.wsdl"

        self.roles = {}
        self.roleShifters = {}
        self.newShifters = {}

        self.diracAdmin = None

    def initialize(self, *args, **kwargs):
        """Initialize."""

        self.lbshiftdburl = self.am_getOption("lbshiftdburl", self.lbshiftdburl)
        self.wsdl = self.am_getOption("wsdl", self.wsdl)

        self.user = self.am_getOption("user")
        self.passwd = self.am_getOption("password")

        if not (self.user and self.passwd):
            self.log.error(f"User and/or password for {self.wsdl} not provided")
            return S_ERROR("Creds not provided")

        self.diracAdmin = DiracAdmin()

        return S_OK()

    def execute(self):
        """Execution."""

        self.roles = {}
        self.roleShifters = {}
        self.newShifters = {}

        self.log.info("Getting roles from CS")
        roles = self.__getRolesFromCS()
        if not roles["OK"]:
            self.log.error(roles["Message"])
            return roles
        self.log.info(f"found {', '.join(roles['Value'])} ")

        self.log.info("Getting role emails")

        for role, eGroup in roles["Value"].items():
            self.roles[role] = eGroup

            email = self.__getRoleEmail(role)
            if not email["OK"]:
                self.log.error(email["Message"])
                # We do not return, we keep execution to clean old shifters
                email["Value"] = None

            email = email["Value"]
            self.roleShifters[eGroup] = (email, role)

            self.log.info(f"{role} -> {email}")

        self.log.info("Setting role emails")
        for eGroup, roleTuple in self.roleShifters.items():
            email, role = roleTuple

            setEmail = self.__setRoleEmail(eGroup, email, role)
            if not setEmail["OK"]:
                self.log.error(setEmail["Message"])

        for newShifterRole, shifterEgroup in self.newShifters.items():
            self.log.info(f"Notifying role {newShifterRole}")
            res = self.__notifyNewShifter(newShifterRole, shifterEgroup)
            if not res["OK"]:
                self.log.error(res["Message"])

        return S_OK()

    def __getRolesFromCS(self):
        """Gets from the CS the roles we want to add to an eGroup.
        Role1 : { eGroup: egroup-blah } in the CS

        returns S_OK( { role1: egroup1, .. } )
        """

        _section = self.am_getModuleParam("section")
        roles = gConfig.getSections(f"{_section}/roles")

        if not roles["OK"]:
            return roles

        eGroups = {}

        for role in roles["Value"]:
            eGroup = gConfig.getValue(f"{_section}/roles/{role}/eGroup")
            if eGroup:
                eGroups[role] = eGroup
                self.log.debug(f"Found {role} : {eGroup} ")

        return S_OK(eGroups)

    def __getRoleEmail(self, role):
        """Get role email from shiftDB."""

        try:
            web = urlopen(self.lbshiftdburl, timeout=60)
        except URLError as e:
            return S_ERROR(f"Cannot open URL: {self.lbshiftdburl}, erorr {e}")

        emailperson = []

        for line in web.readlines():
            for item in json.loads(line):
                if role in item["role"]:
                    # There are three shifts per day, so we take into account what time is it
                    # before sending the email.
                    emailperson.append(item["email"])
                    self.log.info(emailperson)

        if emailperson:
            return S_OK(emailperson)
        else:
            return S_ERROR("Email not found")

    def __setRoleEmail(self, eGroup, email, role):
        """Set email in eGroup."""

        client = suds.client.Client(self.wsdl, username=self.user, password=self.passwd)

        try:
            wgroup = client.service.FindEgroupByName(eGroup)
        except suds.WebFault as wError:
            return S_ERROR(wError)

        members = []
        lastShifterEmail = []
        lastShifterList = {}

        if hasattr(wgroup, "warnings"):
            if wgroup.warnings != []:
                self.log.warn(wgroup.warnings)
        #  return S_ERROR(wgroup.warnings)
        if hasattr(wgroup, "error"):
            if wgroup.error != []:
                self.log.error(wgroup.error)
        if hasattr(wgroup, "result") and hasattr(wgroup.result, "Members"):
            for members in wgroup.result.Members:
                lastShifterEmail.append(members.Email)
                lastShifterList[members.Email] = members

        if email is None:
            self.log.warn("None email. Keeping previous one till an update is found.")
            return S_OK()

        #    if lastShifterEmail.strip().lower() == email.strip().lower():
        #      self.log.info( "%s has not changed as shifter, no changes needed" % email )
        #      return S_OK()

        for lastShifter in lastShifterEmail:
            if email.count(lastShifter) == 0:
                self.log.info(f"{lastShifter} is not anymore shifter, deleting ...")
                try:
                    # The last boolean flag is to overwrite
                    client.service.RemoveEgroupMembers(eGroup, lastShifterList[lastShifter])
                except suds.WebFault as wError:
                    return S_ERROR(wError)

        # Adding a member means it will be the only one in the eGroup, as it is overwritten
        if email != []:
            res = self.__addMember(email, client, eGroup)
            self.log.info(email)
            if not res["OK"]:
                self.log.error(res["Message"])
                return res
            self.log.info(f"{email} added successfully to the eGroup for role {role}")

        self.newShifters[role] = eGroup

        return S_OK()

    def __addMember(self, email, client, wgroup):
        """Adds a new member to the group."""

        # Clear e-Group before inserting anything
        # self.__deleteMembers( client, wgroup )

        self.log.info(f"Adding member {email} to eGroup {wgroup}")

        members = []
        for email_address in email:
            newmember = client.factory.create("ns0:MemberType")
            newmember.Type = "External"
            newmember.Email = email_address

            members.append(newmember)

        try:
            # The last boolean flag is to overwrite
            self.log.info(members)
            client.service.AddEgroupMembers(wgroup, True, members)
        except suds.WebFault as wError:
            return S_ERROR(wError)
        return S_OK()

    def __notifyNewShifter(self, role, eGroup):
        """Sends an email to the shifter ( if any ) at the beginning of the shift
        period."""

        if role == "Production":
            body = __productionBody__
        else:
            self.log.info(f"No email body defined for {role} role")
            return S_OK()

        prodRole = self.roles["Production"]
        geocRole = self.roles["Grid Expert"]
        body = body % (self.roleShifters[prodRole][0], self.roleShifters[geocRole][0])

        # Hardcoded Concezio's email to avoid dirac@mail.cern.ch be rejected by smtp server
        res = self.diracAdmin.sendMail(
            f"{eGroup}@cern.ch", "Shifter information", body, fromAddress="concezio.bozzi@cern.ch"
        )
        return res


__productionBody__ = """Dear GEOC,

this is an (automatic) mail to welcome you on the grid operations shifts.
In order to facilitate your shift activities we wanted to provide you some pointers,
where you could find more information about shifts, the activities and your duties during this period.

http://lhcb-shifters.web.cern.ch/


LHCbDIRAC portal
https://lhcb-portal-dirac.cern.ch/DIRAC


The logbook for LHCb operations, where all activities concerning offline operation are being logged.
https://lblogbook.cern.ch/Operations/
"""



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve an access URL for an LFN replica given a valid DIRAC SE."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    dmScript.registerSiteSwitches()
    Script.registerSwitch("a", "All", "  Also show inactive replicas")
    Script.registerSwitch("", "DiskOnly", "  Show only disk replicas")
    Script.registerSwitch("", "PreferDisk", "  If disk replica, don't show tape replicas")
    Script.registerSwitch("", "ForJobs", "  Select only replicas that can be used for jobs")
    Script.registerSwitch("", "Protocol=", "   Define the protocol for which a tURL is requested (default:root)")
    Script.registerSwitch(
        "",
        "Metalink",
        "   Generate metalink files for parallel download with xrdcp \
                            (one file per lfn). Implies Protocol=root",
    )
    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... [LFN[,LFN2[,LFN3...]]] SE[,SE2...]",
                "Arguments:",
                "  LFN:      Logical File Name or file containing LFNs",
                "  SE:       Valid DIRAC SE",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeAccessURL
    from DIRAC import exit

    exit(executeAccessURL(dmScript))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_Resources_Catalog_BookkeepingDBClient."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
from unittest import mock

import six

from DIRAC import gLogger
import LHCbDIRAC.Resources.Catalog.BookkeepingDBClient as moduleTested

################################################################################


class BookkeepingDBClientt_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""
        gLogger.setLevel("DEBUG")

        # # Mock external libraries / modules not interesting for the unit test
        # mock_pathFinder = mock.Mock()
        # mock_pathFinder.getServiceURL.return_value = 'cookiesURL'
        # self.mock_pathFinder = mock_pathFinder

        # Add mocks to moduleTested
        # moduleTested.PathFinder = self.mock_pathFinder

        mock_RPC = mock.Mock()
        mock_RPC.addFiles.return_value = {"OK": True, "Value": {"Successful": ["A"], "Failed": ["B"]}}
        mock_RPC.removeFiles.return_value = {"OK": True, "Value": {"Successful": ["A"], "Failed": ["B"]}}
        mock_RPC.exists.return_value = {"OK": True, "Value": {"A": 1, "B": 2}}
        mock_RPC.getFileMetadata.return_value = {
            "OK": True,
            "Value": {"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}},
        }
        # mock_RPC.removeMigratingFiles.return_value    = {'OK': True}
        # mock_RPC.removeMigratingReplicas.return_value = {'OK': True}

        mock_BookkeepingClient = mock.Mock()
        mock_BookkeepingClient.return_value = mock_RPC
        self.mock_BookkeepingClient = mock_BookkeepingClient

        moduleTested.BookkeepingClient = self.mock_BookkeepingClient

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.BookkeepingDBClient

    def tearDown(self):
        """TearDown."""
        del self.testClass
        del self.moduleTested
        del self.mock_BookkeepingClient


################################################################################


class BookkeepingDBClient_Success(BookkeepingDBClientt_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""

        catalog = self.testClass()
        self.assertEqual("BookkeepingDBClient", catalog.__class__.__name__)

    def test_init(self):
        """tests that the init method does what it should do."""

        catalog = self.testClass()

        self.assertEqual(1000, catalog.splitSize)
        self.assertEqual("BookkeepingDB", catalog.name)

        catalog = self.testClass()
        self.assertEqual(1000, catalog.splitSize)
        self.assertEqual("BookkeepingDB", catalog.name)

    def test__setHasReplicaFlag(self):
        """test the output of __setHasReplicaFlag."""

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__setHasReplicaFlag([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        mock_RPC = mock.Mock()
        mock_RPC.addFiles.return_value = {"OK": False, "Message": "Bo!"}

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__setHasReplicaFlag([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!", "B": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"C": "Bo!"}}, res["Value"])

        mock_RPC = mock.Mock()
        # side_effect does not work very well, cooked a workaround
        _myValues = [{"OK": False, "Message": "Bo!"}, {"OK": True, "Value": {"A": 1, "B": 2}}]

        def _side_effect(_pfn):
            return _myValues.pop()

        mock_RPC.addFiles.side_effect = _side_effect
        # mock_RPC.addFiles.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
        #                                  {'OK': False, 'Message': 'Bo!'}]

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()
        catalog.splitSize = 2

        res = catalog._BookkeepingDBClient__setHasReplicaFlag([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        _myValues = [{"OK": False, "Message": "Bo!"}, {"OK": True, "Value": {"Successful": ["A"], "Failed": []}}]

        res = catalog._BookkeepingDBClient__setHasReplicaFlag(["A", "C", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "Bo!"}}, res["Value"])

        # Restore the module
        self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
        six.moves.reload_module(self.moduleTested)

    def test__unsetHasReplicaFlag(self):
        """test the output of __unsetHasReplicaFlag."""

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        mock_RPC = mock.Mock()
        mock_RPC.removeFiles.return_value = {"OK": False, "Message": "Bo!"}

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!", "B": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"C": "Bo!"}}, res["Value"])

        mock_RPC = mock.Mock()
        # side_effect does not work very well, cooked a workaround
        _myValues = [{"OK": False, "Message": "Bo!"}, {"OK": True, "Value": {"Successful": [], "Failed": ["A", "B"]}}]

        def _side_effect(_pfn):
            return _myValues.pop()

        mock_RPC.removeFiles.side_effect = _side_effect
        # mock_RPC.removeFiles.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
        #                                     {'OK': False, 'Message': 'Bo!'}]

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()
        catalog.splitSize = 2

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(["A", "C", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "File does not exist", "B": "Bo!"}}, res["Value"])

        # Restore the module
        self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
        six.moves.reload_module(self.moduleTested)

    def test__exists(self):
        """tests the output of __exists."""

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__exists([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__exists(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__exists(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {}}, res["Value"])

        mock_RPC = mock.Mock()
        mock_RPC.exists.return_value = {"OK": False, "Message": "Bo!"}

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__exists([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__exists(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__exists(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!", "B": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__exists(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"C": "Bo!"}}, res["Value"])

        mock_RPC = mock.Mock()

        # side_effect does not work very well, cooked a workaround
        _myValues = [{"OK": False, "Message": "Bo!"}, {"OK": True, "Value": {"A": 1, "B": 2}}]

        def _side_effect(_pfn):
            return _myValues.pop()

        mock_RPC.exists.side_effect = _side_effect
        # mock_RPC.exists.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
        #                                {'OK': False, 'Message': 'Bo!'}]

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()
        catalog.splitSize = 2

        res = catalog._BookkeepingDBClient__exists([])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {}}, res["Value"])

        _myValues = [{"OK": False, "Message": "Bo!"}, {"OK": True, "Value": {"A": 1, "B": 2}}]

        res = catalog._BookkeepingDBClient__exists(["A", "C", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {"B": "Bo!"}}, res["Value"])

        # Restore the module
        self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
        six.moves.reload_module(self.moduleTested)

    def test__getFileMetadata(self):
        """tests the output of __getFileMetadata."""

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__getFileMetadata(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__getFileMetadata(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}, "Failed": {}}, res["Value"])

        res = catalog._BookkeepingDBClient__getFileMetadata(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual(
            {"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}, "Failed": {"C": "File does not exist"}},
            res["Value"],
        )

        mock_RPC = mock.Mock()
        mock_RPC.getFileMetadata.return_value = {"OK": True, "Value": {"Successful": {"A": "1", "B": "2"}}}

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__getFileMetadata(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "1", "B": "2"}}, res["Value"])

        res = catalog._BookkeepingDBClient__getFileMetadata(["C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "1", "B": "2", "C": "File does not exist"}}, res["Value"])

        mock_RPC = mock.Mock()
        mock_RPC.getFileMetadata.return_value = {"OK": False, "Message": "Bo!"}

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__getFileMetadata(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!"}}, res["Value"])

        res = catalog._BookkeepingDBClient__getFileMetadata(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {}, "Failed": {"A": "Bo!", "B": "Bo!"}}, res["Value"])

        mock_RPC = mock.Mock()

        # side_effect does not work very well, cooked a workaround
        _myValues = [{"OK": False, "Message": "Bo!"}, {"OK": True, "Value": {"Successful": {"A": 1, "B": 2}}}]

        def _side_effect(_pfn):
            return _myValues.pop()

        mock_RPC.getFileMetadata.side_effect = _side_effect
        # mock_RPC.getFileMetadata.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
        #                                         {'OK': False, 'Message': 'Bo!'}]

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()
        catalog.splitSize = 2

        res = catalog._BookkeepingDBClient__getFileMetadata(["A", "C", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual(
            {"Successful": {"A": 1, "B": 2}, "Failed": {"C": "File does not exist", "B": "Bo!"}}, res["Value"]
        )

        mock_RPC = mock.Mock()
        mock_RPC.getFileMetadata.return_value = {"OK": True, "Value": {"Successful": {"A": str, "B": "2"}}}

        self.moduleTested.BookkeepingClient.return_value = mock_RPC

        catalog = self.testClass()

        res = catalog._BookkeepingDBClient__getFileMetadata(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": str}, "Failed": {"B": "2"}}, res["Value"])

        # Restore the module
        self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
        six.moves.reload_module(self.moduleTested)

    def test_addFile(self):
        """tests the output of addFile."""

        catalog = self.testClass()

        res = catalog.addFile(1)
        self.assertEqual(False, res["OK"])

        res = catalog.addFile([])
        self.assertEqual(False, res["OK"])

        res = catalog.addFile({})
        self.assertEqual(False, res["OK"])

        res = catalog.addFile(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog.addFile(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog.addFile(["A", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

    def test_addReplica(self):
        """tests the output of addReplica."""

        catalog = self.testClass()

        res = catalog.addReplica(1)
        self.assertEqual(False, res["OK"])

        res = catalog.addReplica([])
        self.assertEqual(False, res["OK"])

        res = catalog.addReplica({})
        self.assertEqual(False, res["OK"])

        res = catalog.addReplica(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog.addReplica(["A"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog.addReplica(["A", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

    def test_removeFile(self):
        """tests the output of removeFile."""

        catalog = self.testClass()

        res = catalog.removeFile(1)
        self.assertEqual(False, res["OK"])

        res = catalog.removeFile([])
        self.assertEqual(False, res["OK"])

        res = catalog.removeFile({})
        self.assertEqual(False, res["OK"])

        res = catalog.removeFile(["A", "B", "C"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        res = catalog.removeFile({"A": 1, "B": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": True}, "Failed": {"B": "File does not exist"}}, res["Value"])

        # FIXME: to be continued, but the method is quite tangled to be tested
        # on a rational way

    def test_removeReplica(self):
        """tests the output of removeReplica."""

        catalog = self.testClass()

        res = catalog.removeReplica(1)
        self.assertEqual(False, res["OK"])

        res = catalog.removeReplica([])
        self.assertEqual(False, res["OK"])

        res = catalog.removeReplica({})
        self.assertEqual(False, res["OK"])

        res = catalog.removeReplica(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.removeReplica(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.removeReplica({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_setReplicaStatus(self):
        """tests the output of setReplicaStatus."""

        catalog = self.testClass()

        res = catalog.setReplicaStatus(1)
        self.assertEqual(False, res["OK"])

        res = catalog.setReplicaStatus([])
        self.assertEqual(False, res["OK"])

        res = catalog.setReplicaStatus({})
        self.assertEqual(False, res["OK"])

        res = catalog.setReplicaStatus(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.setReplicaStatus(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.setReplicaStatus({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_setReplicaHost(self):
        """tests the output of setReplicaHost."""

        catalog = self.testClass()

        res = catalog.setReplicaHost(1)
        self.assertEqual(False, res["OK"])

        res = catalog.setReplicaHost([])
        self.assertEqual(False, res["OK"])

        res = catalog.setReplicaHost({})
        self.assertEqual(False, res["OK"])

        res = catalog.setReplicaHost(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.setReplicaHost(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.setReplicaHost({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_removeDirectory(self):
        """tests the output of removeDirectory."""

        catalog = self.testClass()

        res = catalog.removeDirectory(1)
        self.assertEqual(False, res["OK"])

        res = catalog.removeDirectory([])
        self.assertEqual(False, res["OK"])

        res = catalog.removeDirectory({})
        self.assertEqual(False, res["OK"])

        res = catalog.removeDirectory(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.removeDirectory(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.removeDirectory({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_createDirectory(self):
        """tests the output of createDirectory."""

        catalog = self.testClass()

        res = catalog.createDirectory(1)
        self.assertEqual(False, res["OK"])

        res = catalog.createDirectory([])
        self.assertEqual(False, res["OK"])

        res = catalog.createDirectory({})
        self.assertEqual(False, res["OK"])

        res = catalog.createDirectory(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.createDirectory(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.createDirectory({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_removeLink(self):
        """tests the output of createDirectory."""

        catalog = self.testClass()

        res = catalog.removeLink(1)
        self.assertEqual(False, res["OK"])

        res = catalog.removeLink([])
        self.assertEqual(False, res["OK"])

        res = catalog.removeLink({})
        self.assertEqual(False, res["OK"])

        res = catalog.removeLink(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.removeLink(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.removeLink({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_createLink(self):
        """tests the output of createDirectory."""

        catalog = self.testClass()

        res = catalog.createLink(1)
        self.assertEqual(False, res["OK"])

        res = catalog.createLink([])
        self.assertEqual(False, res["OK"])

        res = catalog.createLink({})
        self.assertEqual(False, res["OK"])

        res = catalog.createLink(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True}, "Failed": {}}, res["Value"])

        res = catalog.createLink(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"path": True, "A": True, "B": True}, "Failed": {}}, res["Value"])

        res = catalog.createLink({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "C": True}, "Failed": {}}, res["Value"])

    def test_exists(self):
        """tests the output of exists."""

        catalog = self.testClass()

        res = catalog.exists(1)
        self.assertEqual(False, res["OK"])

        res = catalog.exists([])
        self.assertEqual(False, res["OK"])

        res = catalog.exists({})
        self.assertEqual(False, res["OK"])

        res = catalog.exists(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {}}, res["Value"])

        res = catalog.exists(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {}}, res["Value"])

        res = catalog.exists({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {}}, res["Value"])

    def test_getFileMetadata(self):
        """tests the output of getFileMetadata."""

        catalog = self.testClass()

        res = catalog.getFileMetadata(1)
        self.assertEqual(False, res["OK"])

        res = catalog.getFileMetadata([])
        self.assertEqual(False, res["OK"])

        res = catalog.getFileMetadata({})
        self.assertEqual(False, res["OK"])

        res = catalog.getFileMetadata(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual(
            {"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}, "Failed": {"path": "File does not exist"}},
            res["Value"],
        )

        res = catalog.getFileMetadata(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual(
            {"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}, "Failed": {"path": "File does not exist"}},
            res["Value"],
        )

        res = catalog.getFileMetadata({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual(
            {"Successful": {"A": {"FileSize": 1}, "B": {"FileSize": 2}}, "Failed": {"C": "File does not exist"}},
            res["Value"],
        )

    def test_getFileSize(self):
        """tests the output of getFileSize."""

        catalog = self.testClass()

        res = catalog.getFileSize(1)
        self.assertEqual(False, res["OK"])

        res = catalog.getFileSize([])
        self.assertEqual(False, res["OK"])

        res = catalog.getFileSize({})
        self.assertEqual(False, res["OK"])

        res = catalog.getFileSize(["path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {"path": "File does not exist"}}, res["Value"])

        res = catalog.getFileSize(["A", "B", "path"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {"path": "File does not exist"}}, res["Value"])

        res = catalog.getFileSize({"A": 2, "C": 3})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"Successful": {"A": 1, "B": 2}, "Failed": {"C": "File does not exist"}}, res["Value"])


#################################################################################



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Uses the DM script switches, and, unless a list of LFNs is provided

1. If --BKQuery is used: get files in BK directories, check if they are in FC
2. If --Production is used get files using the bk query of the given production

Then check if files registered as having a replica in the BK are also in the FC.

If --FixIt is set, take actions

- add files to the BK if they exist in the FC, but have replica = NO in the BK
- set replicaFlag = No in the BK for those files that are not in the FC
"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    # Script initialization
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript
    from DIRAC import gLogger

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [values]",
            ]
        )
    )
    dmScript = DMScript()
    dmScript.registerBKSwitches()
    dmScript.registerFileSwitches()
    Script.registerSwitch("", "FixIt", "   Take action to fix the catalogs")
    Script.registerSwitch("", "CheckAllFlags", "   Consider also files with replica flag NO")
    Script.parseCommandLine(ignoreErrors=True)

    fixIt = False
    checkAll = False
    for opt, val in Script.getUnprocessedSwitches():
        if opt == "FixIt":
            fixIt = True
        elif opt == "CheckAllFlags":
            checkAll = True

    # imports
    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks

    gLogger.setLevel("INFO")
    cc = ConsistencyChecks()
    bkQuery = dmScript.getBKQuery(visible="All")
    cc.bkQuery = bkQuery
    cc.lfns = dmScript.getOption("LFNs", [])
    productions = dmScript.getOption("Productions", [])

    from LHCbDIRAC.DataManagementSystem.Client.CheckExecutors import doCheckBK2FC

    if productions:
        for prod in productions:
            cc.prod = prod
            gLogger.always("Processing production %d" % cc.prod)
            doCheckBK2FC(cc, checkAll, fixIt)
            gLogger.always("Processed production %d" % cc.prod)
    else:
        doCheckBK2FC(cc, checkAll, fixIt)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: TransformationCleaningAgent

.. module: TransformationCleaningAgent

:synopsis: clean up of finalised transformations
"""
import ast

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent as DiracTCAgent

# # from LHCbDIRAC
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

# # agent's name
AGENT_NAME = "Transformation/TransformationCleaningAgent"


class TransformationCleaningAgent(DiracTCAgent):
    """.. class:: TransformationCleaningAgent"""

    def __init__(self, *args, **kwargs):
        """c'tor."""
        DiracTCAgent.__init__(self, *args, **kwargs)

        self.directoryLocations = ["TransformationDB", "StorageUsage"]
        self.archiveAfter = 7
        self.fileTypesToKeep = ["GAUSSHISTMERGED", "LOG"]

        self.bkClient = None
        self.transClient = None
        self.storageUsageClient = None

    #############################################################################

    def initialize(self):
        """Standard initialize method for agents."""
        DiracTCAgent.initialize(self)

        self.directoryLocations = sorted(self.am_getOption("DirectoryLocations", self.directoryLocations))
        self.archiveAfter = self.am_getOption("ArchiveAfter", self.archiveAfter)  # days
        self.fileTypesToKeep = self.am_getOption("FileTypesToKeep", self.fileTypesToKeep)

        self.bkClient = BookkeepingClient()
        self.transClient = TransformationClient()
        self.storageUsageClient = StorageUsageClient()

        return S_OK()

    def cleanMetadataCatalogFiles(self, transID):
        """clean the metadata using BKK and Data Manager. This method is a
        replacement of the one from base class.

        :param self: self reference
        :param int transID: transformation ID
        """
        res = self.bkClient.getProductionFiles(transID, "ALL", "Yes")
        if not res["OK"]:
            return res
        bkMetadata = res["Value"]
        fileToRemove = []
        yesReplica = []
        self.log.info("Found a total of %d files in the BK for transformation %d" % (len(bkMetadata), transID))
        for lfn, metadata in bkMetadata.items():
            if metadata["FileType"] not in self.fileTypesToKeep:
                fileToRemove.append(lfn)
                if metadata["GotReplica"] == "Yes":
                    yesReplica.append(lfn)
        if fileToRemove:
            self.log.info(f"Attempting to remove {len(fileToRemove)} possible remnants from the catalog and storage")
            # Executing with shifter proxy
            gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")
            res = DataManager().removeFile(fileToRemove, force=True)
            gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
            if not res["OK"]:
                return res
            for lfn, reason in res["Value"]["Failed"].items():
                self.log.error("Failed to remove file found in BK", f"{lfn} {reason}")
            if res["Value"]["Failed"]:
                return S_ERROR("Failed to remove all files found in the BK")
            if yesReplica:
                self.log.info(f"Ensuring that {len(yesReplica)} files are removed from the BK")
                res = FileCatalog(catalogs=["BookkeepingDB"]).removeFile(yesReplica)
                if not res["OK"]:
                    return res
                for lfn, reason in res["Value"]["Failed"].items():
                    self.log.error("Failed to remove file from BK", f"{lfn} {reason}")
                if res["Value"]["Failed"]:
                    return S_ERROR("Failed to remove all files from the BK")
        self.log.info("Successfully removed all files found in the BK")
        return S_OK()

    def getTransformationDirectories(self, transID):
        """get the directories for the supplied transformation from the
        transformation system.

        :param self: self reference
        :param int transID: transformation ID
        """

        res = DiracTCAgent.getTransformationDirectories(self, transID)

        if not res["OK"]:
            return res

        directories = res["Value"]
        if isinstance(directories, str):  # Check for (stupid) formats
            directories = ast.literal_eval(directories)
            if not isinstance(directories, list):
                return S_ERROR("Wrong format of output directories")

        if "StorageUsage" in self.directoryLocations:
            res = self.storageUsageClient.getStorageDirectories("", "", transID, [])
            if not res["OK"]:
                self.log.error("Failed to obtain storage usage directories", res["Message"])
                return res
            transDirectories = res["Value"]
            directories = self._addDirs(transID, transDirectories, directories)

        if not directories:
            self.log.info("No output directories found")

        # We should be removing from the list of directories
        # those directories created for file types that are part of those:
        # - uploaded (as output files)
        # - not merged by subsequent steps
        # but this is pretty difficult to identify at run time, so we better remove the "RemovingFiles" production status
        # and replace it with a flush (this applies only to MC).
        # So we just have a created list.
        fileTypesToKeepDirs = []
        for fileTypeToKeep in self.fileTypesToKeep:
            fileTypesToKeepDirs.extend([x for x in directories if fileTypeToKeep in x])
        directories = list(set(directories).difference(set(fileTypesToKeepDirs)))

        directories = sorted(directories)
        return S_OK(directories)



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import sys
from enum import StrEnum

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise


class TableJoins(StrEnum):
    """Enum class to store the table joins for the query.

    This class is used to store the table joins for when using the files table.
    The order of the values in the enum is important as it is used to order
    the joins in the query to ensure dependent tables are joined first.
    """

    JOBS = "JOIN jobs j ON f.jobid = j.jobid AND f.production = j.production"
    PRODUCTIONSCONTAINER = "JOIN productionscontainer cont ON f.production = cont.production"
    CONFIGURATIONS = "JOIN configurations c ON cont.configurationid = c.configurationid"
    FILETYPES = "JOIN filetypes ft ON f.filetypeid = ft.filetypeid"
    DATAQUALITY = "JOIN dataquality d ON f.qualityid = d.qualityid"
    CONDESC = "JOIN condesc cd ON (cont.simid = cd.simid OR cont.daqperiodid = cd.daqid)"
    PROCPATHS = "JOIN procpaths pp ON pp.id = cont.processingid"
    RUNS = "JOIN runs ON runs.runnumber = j.runnumber"
    SMOG2STATE = "JOIN smog2 s2s ON runs.smog2_id = s2s.id"

    TMP_PRODUCTION = "JOIN LHCB_DIRACBOOKKEEPING.temp_query_production tmp_p ON f.production = tmp_p.production"
    TMP_FILETYPES = "JOIN LHCB_DIRACBOOKKEEPING.temp_query_filetype tmp_ft ON ft.name = tmp_ft.name"

    @staticmethod
    def sort_key(value: str) -> int:
        try:
            return list(TableJoins).index(value)
        except ValueError:
            return sys.maxsize


TABLE_JOIN_TO_NAME = {
    None: "f",
    TableJoins.JOBS: "j",
    TableJoins.CONFIGURATIONS: "c",
    TableJoins.PRODUCTIONSCONTAINER: "cont",
    TableJoins.FILETYPES: "ft",
    TableJoins.DATAQUALITY: "d",
    TableJoins.CONDESC: "cd",
    TableJoins.PROCPATHS: "pp",
    TableJoins.RUNS: "runs",
    TableJoins.SMOG2STATE: "s2s",
}
JOIN_TO_TEMP_TABLE = {
    TableJoins.TMP_PRODUCTION: "temp_query_production",
    TableJoins.TMP_FILETYPES: "temp_query_filetype",
}

CTE_PROCPATHS_FULL = (
    "procpaths (procpath, id) as (\n"
    "    SELECT distinct SUBSTR(SYS_CONNECT_BY_PATH(name, '/'), 2), id\n"
    "    FROM processing\n"
    "    START WITH id in (select distinct id from processing where parentid is NULL)\n"
    "    CONNECT BY NOCYCLE PRIOR id=parentid\n"
    ")"
)
# If we're filtering on :procpath split it by / and only keep potentially
# relavant rows from the processing table to reduce the query's cost
CTE_PROCPATHS_FILTERED = (
    "procpaths (procpath, id) AS (\n"
    "    SELECT SUBSTR(SYS_CONNECT_BY_PATH(name, '/'), 2) as procpath, id\n"
    "    FROM (\n"
    "        select * from processing where name in (\n"
    "            SELECT REGEXP_SUBSTR(:procPass, '[^/]+', 1, LEVEL)\n"
    "            FROM dual\n"
    "            CONNECT BY REGEXP_SUBSTR(:procPass, '[^/]+', 1, LEVEL) IS NOT NULL\n"
    "        )\n"
    "    )\n"
    "    START WITH id IN (SELECT DISTINCT id FROM processing WHERE name = REGEXP_SUBSTR(:procPass, '[^/]+', 1, 1))\n"
    "    CONNECT BY NOCYCLE PRIOR id = parentid\n"
    ")"
)
CTE_CONDESC = (
    "condesc (daqid, simid, description) as (\n"
    "    SELECT distinct DAQPERIODID, NULL, DESCRIPTION FROM data_taking_conditions\n"
    "    UNION\n"
    "    SELECT distinct NULL, simid, simdescription FROM simulationConditions\n"
    ")"
)


def field_condition_builder_via_temp_table(
    join: TableJoins,
    values: str | int | list[str | int],
    pre_inserts: list[tuple[str, list[tuple[str]]]],
    joins: set[TableJoins],
) -> None:
    """Helper function to build a condition for a column with a single value or a list of values.

    This function adds an entry to the pre_inserts list to insert the values into a temporary table.
    The function also adds the corresponding join to the set of joins.
    """
    insert_stmt = f"INSERT INTO LHCB_DIRACBOOKKEEPING.{JOIN_TO_TEMP_TABLE[join]} VALUES (:1)"
    if isinstance(values, (str, int)):
        values = [values]
    pre_inserts.append((insert_stmt, [(value,) for value in values]))
    joins.add(join)


def field_condition_builder_via_params(
    column: str,
    value: str | int | list[str | int],
    param_prefix: str,
    kwparams: dict[str, str | int],
    conditions: set[str],
) -> None:
    """Helper function to build a condition for a column with a single value or a list of values.

    The function adds the condition to the set of conditions and updates the kwparams dictionary with the
    appropriate parameters. Parameters are named with the prefix followed by a number, starting from 0 for
    the first value in the list.

    The prefix must be unique for each column, as the function will overwrite the parameters in the kwparams.
    """
    if isinstance(value, (str, int)):
        conditions.add(f"{column} = :{param_prefix}")
        kwparams[param_prefix] = value
    elif isinstance(value, (list, tuple)):
        if len(set(value)) > 100:
            raise NotImplementedError(f"Too many values for IN clause on {column}, should use temp table")
        parameters = {f"{param_prefix}{j}": prod for j, prod in enumerate(value)}
        kwparams.update(parameters)
        conditions.add(f"{column} IN ({', '.join(f':{x}' for x in parameters)})")
    else:
        raise TypeError(f"Invalid type for {column}: {type(value)}")


def combineDescription(simdesc, datataking):
    """Combine the simulation description and data taking description.

    TODO: The calling code should be refactored to remove the separate handling of simdesc and datataking.
    """
    if simdesc is None or simdesc == "ALL":
        return datataking
    if datataking is None or datataking == "ALL":
        return None
    if simdesc != datataking:
        raise ValueError(f"Conditions inconsistency: {simdesc=} {datataking=}")
    return datataking


def buildQuery(
    dbR,
    selection,
    conddescription,
    procPass,
    ftype,
    evt,
    configName,
    configVersion,
    production,
    flag,
    startDate,
    endDate,
    startRunID,
    endRunID,
    runnumbers,
    replicaFlag,
    visible,
    tcks,
    jobStart,
    jobEnd,
    smog2States,
    dqok,
    seed_md5,
    sample_max,
    *,
    forceJoins=None,
    extraConditions=None,
):
    joins = set()
    ctes = set()
    pre_inserts = []
    if forceJoins:
        joins.update(forceJoins)
    conditions = set()
    if extraConditions:
        conditions.update(extraConditions)
    kwparams = {}

    # TODO: When checking for all we should apply str.upper() if it's a string
    if configName is not None and configName != "ALL":
        joins.add(TableJoins.PRODUCTIONSCONTAINER)
        joins.add(TableJoins.CONFIGURATIONS)
        conditions.add("c.configname = :configName")
        kwparams["configName"] = configName

    if configVersion is not None and configVersion != "ALL":
        joins.add(TableJoins.PRODUCTIONSCONTAINER)
        joins.add(TableJoins.CONFIGURATIONS)
        conditions.add("c.configversion = :configVersion")
        kwparams["configVersion"] = configVersion

    if production is not None and production != "ALL":
        joins.add(TableJoins.PRODUCTIONSCONTAINER)
        field_condition_builder_via_temp_table(TableJoins.TMP_PRODUCTION, production, pre_inserts, joins)

    if tcks is not None and tcks != "ALL" and tcks:
        joins.add(TableJoins.JOBS)
        field_condition_builder_via_params("j.tck", tcks, "tck", kwparams, conditions)

    if procPass is not None and procPass != "ALL":
        procPass = procPass.lstrip("/")
        joins.add(TableJoins.PRODUCTIONSCONTAINER)
        joins.add(TableJoins.PROCPATHS)
        ctes.add(CTE_PROCPATHS_FILTERED)
        # TODO: We probably need to support like here for some other methods
        conditions.add("pp.procpath = :procPass")
        kwparams["procPass"] = procPass

    if ftype is not None and ftype != "ALL" and ftype:
        joins.add(TableJoins.FILETYPES)
        field_condition_builder_via_temp_table(TableJoins.TMP_FILETYPES, ftype, pre_inserts, joins)

    if runnumbers is not None and runnumbers != "ALL" and runnumbers:
        # As far as I can tell LegacyOracleBookkeepingDB has bizarre behavior
        # if both runnumbers and startRunID/endRunID are specified so let's
        # see if we can get away with not supporting both.
        if startRunID is not None and startRunID != "ALL":
            raise NotImplementedError("Both runnumbers and startRunID are specified")
        if endRunID is not None and endRunID != "ALL":
            raise NotImplementedError("Both runnumbers and endRunID are specified")
        joins.add(TableJoins.JOBS)
        field_condition_builder_via_params("j.runnumber", runnumbers, "run", kwparams, conditions)

    if startRunID is not None and startRunID != "ALL":
        joins.add(TableJoins.JOBS)
        conditions.add("j.runnumber >= :startRunID")
        kwparams["startRunID"] = startRunID

    if endRunID is not None and endRunID != "ALL":
        joins.add(TableJoins.JOBS)
        conditions.add("j.runnumber <= :endRunID")
        kwparams["endRunID"] = endRunID

    if evt != 0 and evt is not None and evt != "ALL" and evt:
        field_condition_builder_via_params("f.eventtypeid", evt, "evt", kwparams, conditions)

    if startDate is not None and startDate != "ALL":
        conditions.add("f.inserttimestamp >= TO_TIMESTAMP (:startDate,'YYYY-MM-DD HH24:MI:SS')")
        kwparams["startDate"] = startDate

    # TODO __buildStartenddate uses utcnow for some reason, seems useless but someone else should check
    if endDate is not None and endDate != "ALL":
        conditions.add("f.inserttimestamp <= TO_TIMESTAMP (:endDate,'YYYY-MM-DD HH24:MI:SS')")
        kwparams["endDate"] = endDate

    if jobStart is not None and jobStart != "ALL":
        joins.add(TableJoins.JOBS)
        conditions.add("j.jobstart >= TO_TIMESTAMP (:jobStart,'YYYY-MM-DD HH24:MI:SS')")
        kwparams["jobStart"] = jobStart

    if jobEnd is not None and jobEnd != "ALL":
        joins.add(TableJoins.JOBS)
        conditions.add("j.jobend <= TO_TIMESTAMP (:jobEnd,'YYYY-MM-DD HH24:MI:SS')")
        kwparams["jobEnd"] = jobEnd

    if flag is not None and flag != "ALL" and flag:
        joins.add(TableJoins.DATAQUALITY)
        field_condition_builder_via_params("d.dataqualityflag", flag, "dq", kwparams, conditions)

    if dqok:
        joins.add(TableJoins.JOBS)
        if flag is None or flag == "ALL":
            raise ValueError("DataQuality OK must be explicitly specified to use ExtendedDQOK")
        if isinstance(flag, str):
            flag = [flag]
        if flag != ["OK"]:
            raise ValueError("ExtendedDQOK can be specified with DataQuality=OK only")
        if isinstance(dqok, str):
            dqok = [dqok]
        for i, system in enumerate(set(dqok)):
            joins.add(f"JOIN extendeddqok dqok{i} ON dqok{i}.runnumber = j.runnumber")
            conditions.add(f"dqok{i}.systemname = :dqok{i}")
            kwparams[f"dqok{i}"] = system

    if replicaFlag is not None and replicaFlag != "ALL":
        conditions.add("f.gotreplica = :replicaFlag")
        kwparams["replicaFlag"] = replicaFlag

    if visible is not None and visible != "ALL":
        conditions.add("f.visibilityflag = :visibilityflag")
        kwparams["visibilityflag"] = visible[0].upper()

    if conddescription is not None and conddescription != "ALL":
        joins.add(TableJoins.PRODUCTIONSCONTAINER)
        joins.add(TableJoins.CONDESC)
        ctes.add(CTE_CONDESC)
        conditions.add("cd.description = :conddescription")
        kwparams["conddescription"] = conddescription

    if smog2States is not None and smog2States != "ALL":
        joins.add(TableJoins.RUNS)
        joins.add(TableJoins.SMOG2STATE)
        if isinstance(smog2States, str):
            smog2States = [smog2States]
        smog2States = set(smog2States)
        smog2Conditions = []
        if "Undefined" in smog2States:
            smog2States.remove("Undefined")
            smog2Conditions.append("s2s.id IS NULL")
        if smog2States:
            stateVars = []
            for i, state in enumerate(smog2States):
                kwparams[f"smog2State{i}"] = state
                stateVars.append(f":smog2State{i}")
            smog2Conditions.append(f"s2s.state IN ({', '.join(stateVars)})")
        if len(smog2Conditions) < 2:
            conditions.add(smog2Conditions[0])
        else:
            conditions.add("(" + " OR ".join(smog2Conditions) + ")")

    if seed_md5 or sample_max:
        if not (seed_md5 and sample_max):
            raise ValueError("Both seed_md5 and sample_max MUST be specified.")

        if not (len(seed_md5) == len(sample_max) == 32 and seed_md5.isupper() and sample_max.isupper()):
            raise ValueError("seed_md5 and sample_max must each be 32-character upper-case MD5 hashes.")

        md5_hash_stmt = "STANDARD_HASH(CONCAT(f.filename, :seed_md5), 'MD5')"
        conditions.add(f"({md5_hash_stmt} <= :sample_max)")

        kwparams["seed_md5"] = seed_md5
        kwparams["sample_max"] = sample_max

    query = []
    if ctes:
        query.append("WITH " + ", ".join(ctes))
    query += [
        f"SELECT {selection}",
        "FROM files f",
    ]
    query += sorted(joins, key=TableJoins.sort_key)
    if conditions:
        query.append(f"WHERE {' AND '.join(sorted(conditions))}")
    return returnValueOrRaise(dbR.query("\n".join(query), kwparams=kwparams, pre_inserts=pre_inserts))



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Check if all files have a replica in a certain (set of) SE )Tier1-Archive
default) List the files that don't have a replica in the specified SE (group)"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    # Script initialization
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    Script.registerSwitch("", "FixIt", "   Take action to fix the catalogs")
    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [values]",
            ]
        )
    )
    dmScript = DMScript()
    dmScript.registerDMSwitches()  # Directory
    Script.parseCommandLine(ignoreErrors=True)
    fixIt = False
    for opt, val in Script.getUnprocessedSwitches():
        if opt == "FixIt":
            fixIt = True

    # imports
    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks

    cc = ConsistencyChecks()
    cc.directories = dmScript.getOption("Directory", [])
    cc.lfns = dmScript.getOption("LFNs", []) + [lfn for arg in Script.getPositionalArgs() for lfn in arg.split(",")]
    bkQuery = dmScript.getBKQuery(visible="All")
    if bkQuery.getQueryDict() != {"Visible": "All"}:
        bkQuery.setOption("ReplicaFlag", "All")
        cc.bkQuery = bkQuery
    seList = dmScript.getOption("SEs", [])
    if not seList:
        dmScript.setSEs("Tier1-Archive")
        seList = dmScript.getOption("SEs", [])

    from LHCbDIRAC.DataManagementSystem.Client.CheckExecutors import doCheckSE

    doCheckSE(cc, seList, fixIt)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get Bookkeeping paths given a decay.

@author Vanya BELYAEV Ivan.Belyaev@itep.ru
        Federico Stagni fstagni@cern.ch
"""
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} eventType  "]))

    Script.registerSwitch("p", "production", "Obtain the paths in ``Production format'' for Ganga")
    Script.parseCommandLine(ignoreErrors=True)

    productionFormat = False
    for p, _v in Script.getUnprocessedSwitches():
        if p.lower() in ("p", "production"):
            productionFormat = True
        break

    args = Script.getPositionalArgs()
    if len(args) < 1:
        Script.showHelp(exitCode=1)

    eventType = args[0]

    bkClient = BookkeepingClient()

    # # get productions for given event type
    res = bkClient.getProductionSummaryFromView({"EventType": eventType, "Visible": True})
    if not res["OK"]:
        gLogger.error(f"Could not retrieve production summary for event {eventType}", res["Message"])
        DIRAC.exit(1)

    # # get production-IDs
    prodIDs = [p["Production"] for p in res["Value"]]

    # # loop over all productions
    for prodID in sorted(prodIDs):
        res = bkClient.getProductionInformation(prodID)
        if not res["OK"]:
            gLogger.error(f"Could not retrieve production infos for production {prodID}", res["Message"])
            continue
        prodInfo = res["Value"]
        steps = prodInfo["Steps"]
        if isinstance(steps, str):
            continue
        files = prodInfo["Number of files"]
        events = prodInfo["Number of events"]
        path = prodInfo["Path"]
        dddb = prodInfo["Steps"][0][4]
        conddb = prodInfo["Steps"][0][5]

        evts = 0
        ftype = None
        for i in events:
            if i[0] in ["GAUSSHIST", "LOG", "SIM", "DIGI", "RAW"]:
                continue
            # NB of events can sometimes be None
            if i[1] is not None:
                evts += i[1]

            if not ftype:
                ftype = i[0]

        nfiles = 0
        for f in files:
            if f[1] in ["GAUSSHIST", "LOG", "SIM", "DIGI", "RAW"]:
                continue
            if f[1] != ftype:
                continue
            nfiles += f[0]

        p0, n, p1 = path.partition("\n")
        if n:
            path = p1

        if productionFormat:
            p, s, e = path.rpartition("/")
            if s and e:
                path = "/%d/%d/%s" % (prodID, int(eventType), e)
        print(path, dddb, conddb, nfiles, evts, prodID)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Generic XMLParser used to convert a XML file into more pythonic style.

A tree ox XMLNodes.
"""
import xml.dom.minidom

################################################################################


class XMLNode:
    """XMLNodes represent XML elements.

    May have attributes. They have either children or value ( exclusive-
    or).
    """

    def __init__(self, name):
        self.name = name
        self.attributes = {}
        self.children = None
        self.value = None

    def childrens(self, name):
        """return children."""
        return [child for child in self.children if child.name == name]

    def __repr__(self):
        return f"< {self.name} >"


################################################################################


class XMLTreeParser:
    """XMLTreeParser converts an XML file or a string into a tree of XMLNodes. It
    does not validate the XML.

    Elements that are the only child of an element and are of text or
    cdata type are considered to be the value of their parent.
    """

    def __init__(self):
        self.tree = None

    def parse(self, xmlFile):
        """parse the XML."""
        domXML = xml.dom.minidom.parse(xmlFile)
        self.__handleXML(domXML)
        return self.tree

    def parseString(self, xmlString):
        """parse the XML."""
        domXML = xml.dom.minidom.parseString(xmlString)
        self.__handleXML(domXML)
        return self.tree

    ################################################################################
    # AUXILIAR FUNCTIONS
    ################################################################################

    def __handleXML(self, domXML):
        """handles first child."""
        self.tree = self.__handleElement([domXML.firstChild])

    def __handleElement(self, elements):
        """treat each element."""
        nodes = []

        for el in elements:
            if el.nodeType == el.TEXT_NODE:
                continue

            node = XMLNode(el.localName)
            node.attributes = self.__getAttributesDict(el)

            if len(el.childNodes) == 1:
                childNode = el.childNodes[0]
                if childNode.nodeType == childNode.TEXT_NODE or childNode.nodeType == childNode.CDATA_SECTION_NODE:
                    node.value = self.__handleTextElement(el.childNodes[0])
                else:
                    node.children = self.__handleElement(el.childNodes)
            else:
                node.children = self.__handleElement(el.childNodes)

            nodes.append(node)

        return nodes

    @staticmethod
    def __getAttributesDict(element):
        """get the attributes in a dictionary."""
        dictionary = {}
        if element.attributes:
            for attr in element.attributes.values():
                dictionary[attr.name] = attr.value
        return dictionary

    def __handleTextElement(self, textElement):
        """treat the Text element."""
        return self.__getText(textElement)

    @staticmethod
    def __getText(node):
        """get the TEXT."""
        data = ""
        if node.nodeType == node.TEXT_NODE or node.nodeType == node.CDATA_SECTION_NODE:
            data = node.data
        return data



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""StepAccounting module performs several common operations at the end of a
workflow step, in particular prepares and sends the step accounting data."""
import datetime

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.Workflow.Utilities.Utils import getStepCPUTimes

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.AccountingSystem.Client.Types.JobStep import JobStep


class StepAccounting(ModuleBase):
    """StepAccounting class."""

    def __init__(self, bkClient=None, dm=None):
        self.log = gLogger.getSubLogger("StepAccounting")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.dsc = None
        self.stepStat = None

    ########################################################################

    def _resolveInputVariables(self, dsc=None):
        """By convention all workflow parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

        if dsc is not None:
            self.dsc = dsc
        else:
            self.dsc = self.workflow_commons["AccountingReport"]

        if self.stepStatus["OK"]:
            self.stepStat = "Done"
        else:
            self.stepStat = "Failed"

    ########################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
        js=None,
        xf_o=None,
        dsc=None,
    ):
        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            # Check if the step is worth accounting
            if "applicationName" not in self.step_commons:
                self.log.info("Not an application step: it will not be accounted")
                return S_OK()

            ########################################################################
            # Timing
            execTime, cpuTime = getStepCPUTimes(self.step_commons)
            normCPU = cpuTime
            cpuNormFactor = gConfig.getValue("/LocalSite/CPUNormalizationFactor", 0.0)
            if cpuNormFactor:
                normCPU = cpuTime * cpuNormFactor

            if not js:
                jobStep = JobStep()
            else:
                jobStep = js

            if not xf_o:
                try:
                    xf_o = self.step_commons["XMLSummary_o"]
                except KeyError:
                    self.log.error("XML Summary object could not be found (not produced?), skipping the report")
                    return S_OK()

            self._resolveInputVariables(dsc)

            now = datetime.datetime.utcnow()
            jobStep.setStartTime(now)
            jobStep.setEndTime(now)

            dataDict = {
                "JobGroup": str(self.production_id),
                "RunNumber": self.runNumber,
                "EventType": self.eventType,
                "ProcessingType": self.stepProcPass,  # this is the processing pass of the step
                "ProcessingStep": self.BKstepID,  # the step ID
                "Site": self.siteName,
                "FinalStepState": self.stepStat,
                "CPUTime": cpuTime,
                "NormCPUTime": normCPU,
                "ExecTime": execTime * self.numberOfProcessors,
                "InputData": sum(xf_o.inputFileStats.values()),
                "OutputData": sum(xf_o.outputFileStats.values()),
                "InputEvents": xf_o.inputEventsTotal,
                "OutputEvents": xf_o.outputEventsTotal,
            }

            jobStep.setValuesFromDict(dataDict)

            res = jobStep.checkValues()
            if not res["OK"]:
                self.log.error("Values for StepAccounting are wrong", res["Message"], dataDict)
                return S_ERROR("Values for StepAccounting are wrong")

            if not self._enableModule():
                self.log.info(f"Not enabled, would have accounted for {dataDict}")
                return S_OK()

            self.dsc.addRegister(jobStep)

            return S_OK()

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in StepAccounting execute module", lException=e)
            return S_ERROR(str(e))

        finally:
            super().finalize()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Returns the platform supported by the current WN."""
import LbPlatformUtils
from DIRAC.Core.Base.Script import Script


def sendMail(msg=""):
    """send a notification mail when no platform is found."""
    from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
    from DIRAC import gConfig

    mailAddress = Operations().getValue("EMail/JobFailures", "Vladimir.Romanovskiy@cern.ch")
    site = gConfig.getValue("LocalSite/Site")
    ce = gConfig.getValue("LocalSite/GridCE")
    queue = gConfig.getValue("LocalSite/CEQueue")
    body = "*** THIS IS AN AUTOMATED MESSAGE ***" + "\n\n" + msg + "\n\n"
    body = body + f"At site {site}, CE = {ce}, queue = {queue}" + "\n\n"

    for mA in mailAddress.replace(" ", "").split(","):
        NotificationClient().sendMail(
            mailAddress, "Problem with DIRAC architecture", body, "federico.stagni@cern.ch", localAttempt=False
        )


@Script()
def main():
    Script.registerSwitch("", "BinaryTag", "   Print the host binary tag instead of the host dirac_platform")
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import gConfig, gLogger, exit as dExit
    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

    parList = Script.getUnprocessedSwitches()
    for switch, _val in parList:
        if switch == "BinaryTag":
            try:
                # Get the binaryTag name. If an error occurs, an exception is thrown
                binaryTag = LbPlatformUtils.host_binary_tag()
                if not binaryTag:
                    gLogger.fatal("There is no binaryTag corresponding to this machine")
                    sendMail("There is no binaryTag corresponding to this machine")
                    dExit(1)
                print(binaryTag)
                dExit(0)
            except Exception as e:
                msg = "Exception getting binaryTag: " + repr(e)
                gLogger.exception(msg, lException=e)
                sendMail(msg)
                dExit(1)

    try:
        site = gConfig.getValue("/LocalSite/Site")
        grid = site.split(".")[0]
        ce = gConfig.getValue("/LocalSite/GridCE")

        allowContainers = gConfig.getValue(f"/Resources/Sites/{grid}/{site}/CEs/{ce}/AllowContainers", None)
        if allowContainers is None:
            allowContainers = gConfig.getValue(f"/Resources/Sites/{grid}/{site}/AllowContainers", None)
        if allowContainers is None:
            allowContainers = Operations().getValue("GaudiExecution/AllowContainers", "no")

        if allowContainers.lower() in ("yes", "true", "all"):
            allowContainers = True
        elif allowContainers.lower() in ("no", "false", "none", ""):
            allowContainers = False
        else:
            gLogger.warn("Invalid value for AllowContainers", repr(allowContainers))
            allowContainers = False

        # Get the platform name. If an error occurs, an exception is thrown
        platform = LbPlatformUtils.dirac_platform(allow_containers=allowContainers)
        if not platform:
            gLogger.fatal("There is no platform corresponding to this machine")
            sendMail("There is no platform corresponding to this machine")
            dExit(1)
        print(platform)
        dExit(0)

    except Exception as e:
        msg = "Exception getting platform: " + repr(e)
        gLogger.exception(msg, lException=e)
        sendMail(msg)
        dExit(1)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""NotifyAgent This agent reads a cache file ( cache.db ) which contains the
aggregated information of what happened to each production request. After
reading the cache file ( by default every 30 minutes ) it sends an email for
every site and then clears it.

Please note that this agent is a hybrid agent that sends aggregated emails for both
LHCbDIRAC.ProductionManagementSystem.Utilities.Utils.informPeople and
LHCbDIRAC.ProductionManagementSystem.Agent.ProductionStatusAgent._mailProdManager
"""
import os
import sqlite3

import DIRAC
from DIRAC import S_OK, gConfig
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
from LHCbDIRAC.ProductionManagementSystem.Utilities.Utils import _getMemberMails

AGENT_NAME = "ProductionManagement/NotifyAgent"


class NotifyAgent(AgentModule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.diracAdmin = None
        self.csS = None
        self.fromAddress = None
        self.cacheFile = os.path.join(DIRAC.rootPath, "work/ProductionManagement/cache.db")
        self.statusUpdateMail = False

    def initialize(self):
        """NotifyAgent initialization."""

        try:
            with sqlite3.connect(self.cacheFile) as conn:
                conn.execute(
                    """CREATE TABLE IF NOT EXISTS ProductionManagementCache(
                      reqId VARCHAR(64) NOT NULL DEFAULT "",
                      reqType VARCHAR(64) NOT NULL DEFAULT "",
                      reqWG VARCHAR(64) NOT NULL DEFAULT "",
                      reqName VARCHAR(64) NOT NULL DEFAULT "",
                      SimCondition VARCHAR(64) NOT NULL DEFAULT "",
                      ProPath VARCHAR(64) NOT NULL DEFAULT "",
                      thegroup VARCHAR(64) NOT NULL DEFAULT "",
                      reqInform VARCHAR(64) NOT NULL DEFAULT ""
                     );"""
                )
        except sqlite3.OperationalError:
            self.log.error("Email cache database is locked")

        self.diracAdmin = DiracAdmin()

        self.csS = PathFinder.getServiceSection("ProductionManagement", "ProductionRequest")

        self.fromAddress = gConfig.getValue(f"{self.csS}/fromAddress", "")
        if not self.fromAddress:
            self.log.info("No fromAddress is defined, a default value will be used instead")
            self.fromAddress = "vladimir.romanovsky@cern.ch"
        self.statusUpdateMail = self.am_getOption("StatusUpdateMail", self.statusUpdateMail)

        return S_OK()

    def execute(self):
        if not os.path.isfile(self.cacheFile):
            self.log.error(self.cacheFile + " does not exist.")
            return S_OK()

        with sqlite3.connect(self.cacheFile) as conn:
            if not self.csS:
                self.log.error("No ProductionRequest section in configuration")
                return S_OK()
            self._executeForProductionManagementSystem(conn)
            self._executeForProductionStatusAgent(conn)

        return S_OK()

    def _executeForProductionManagementSystem(self, conn):
        """This is for the ProductionManagementSystem's Utilities"""
        aggregated_body = """\
          <!DOCTYPE html>
          <html>
          <head>
          <meta charset='UTF-8'>
            <style>
              table{color:#333;font-family:Helvetica,Arial,sans-serif;min-width:850px;border-collapse:collapse;border-spacing:0}
              td,th{border:1px solid transparent;height:30px;transition:all .3s}th{background:#DFDFDF;font-weight:700}
              td{background:#FAFAFA;text-align:center}tr:nth-child(even) td{background:#F1F1F1}tr:nth-child(odd)
              td{background:#FEFEFE}tr td:hover{background:#666;color:#FFF}tr td.link:hover{background:inherit;}
              p{width: 850px;}
            </style>
          </head>
          <body>
          """

        result = conn.execute(
            "SELECT DISTINCT thegroup, reqName, reqWG, reqInform, reqType from ProductionManagementCache;"
        )
        for thegroup, reqName, reqWG, reqInform, reqType in result:
            link = (
                "https://lhcb-portal-dirac.cern.ch/DIRAC/s:"
                + PathFinder.getDIRACSetup()
                + "/g:"
                + thegroup
                + "/?view=tabs&url_state=1|*LHCbDIRAC.ProductionRequestManager.classes.ProductionRequestManager:"
            )

            aggregated_body = ""
            html_elements = ""

            # Skip if group is empty
            if not thegroup:
                continue
            # Only ask people to act on MC requests
            if reqType != "Simulation":
                continue

            if thegroup == "lhcb_bk":
                header = (
                    "New Productions are requested and they have customized Simulation Conditions. "
                    "As member of <span style='color:green'>" + thegroup + "</span> group, you are asked either to "
                    "register new Simulation conditions or to reject the requests. In case some other member of the "
                    "group has already done that, please ignore this mail.\n"
                )

            elif thegroup == "lhcb_ppg":
                header = (
                    "New Productions are requested. As member of <span style='color:green'>" + thegroup + "</span> "
                    "group, you are asked either to sign or to reject it. In case some other member of the group has "
                    "already done that, please ignore this mail.\n"
                )
            else:
                header = (
                    "As member of <span style='color:green'>"
                    + thegroup
                    + "</span> group, your are asked to review the below requests.\n"
                )

            cursor = conn.execute(
                "SELECT reqId, reqType, reqWG, reqName, SimCondition, ProPath from ProductionManagementCache "
                "WHERE thegroup = ? and reqName=? and reqWG=? and reqType=? ",
                (thegroup, reqName, reqWG, reqType),
            )

            for row_reqId, row_reqType, row_reqWG, row_reqName, row_SimCondition, row_ProPath in cursor:
                html_elements += (
                    "<tr>"
                    + "<td>"
                    + row_reqId
                    + "</td>"
                    + "<td>"
                    + row_reqName
                    + "</td>"
                    + "<td>"
                    + row_reqType
                    + "</td>"
                    + "<td>"
                    + row_reqWG
                    + "</td>"
                    + "<td>"
                    + row_SimCondition
                    if row_SimCondition
                    else (
                        "" + "</td>" + "<td>" + row_ProPath
                        if row_ProPath
                        else ""
                        + "</td>"
                        + "<td class='link'><a href='"
                        + link
                        + "' target='_blank'> Link </a></td>"
                        + "</tr>"
                    )
                )

            # If there are no requests to display, don't bother sending emails
            if not html_elements:
                continue

            aggregated_body += f"""        <p>{header}</p>
        <table>
          <tr>
              <th>ID</th>
              <th>Name</th>
              <th>Type</th>
              <th>Working Group</th>
              <th>Conditions</th>
              <th>Processing pass</th>
              <th>Link</th>
          </tr>
          {html_elements}
        </table>
      </body>
      </html>
      """

            if reqInform:
                for emailaddress in reqInform.split(","):
                    res = self.diracAdmin.sendMail(
                        emailaddress,
                        f"Notifications for production requests - Group {thegroup}; {reqWG}; {reqName}",
                        aggregated_body,
                        self.fromAddress,
                        html=True,
                    )

            for people in _getMemberMails(thegroup):
                res = self.diracAdmin.sendMail(
                    people,
                    f"Notifications for production requests - Group {thegroup}; {reqWG}; {reqName}",
                    aggregated_body,
                    self.fromAddress,
                    html=True,
                )

                if res["OK"]:
                    conn.execute("DELETE FROM ProductionManagementCache;")
                else:
                    self.log.error(f"_inform_people: can't send email: {res['Message']}")

    def _executeForProductionStatusAgent(self, conn):
        """This is for the ProductionStatusAgent"""

        if self.statusUpdateMail:
            aggregated_body = """\
            <!DOCTYPE html>
            <html>
            <head>
            <meta charset='UTF-8'>
                <style>
                table{color:#333;font-family:Helvetica,Arial,sans-serif;min-width:700px;border-collapse:collapse;border-spacing:0}
                td,th{border:1px solid transparent;height:30px;transition:all .3s}th{background:#DFDFDF;font-weight:700}
                td{background:#FAFAFA;text-align:center}.setup{font-size:150%;color:grey}.Active{color:green}
                .Archived,.Cleaned,.Cleaning{color:gray}.Completed{color:purple}.Idle{color:#90ee90}.Stopped{color:orange}
                .Testing,.TransformationCleaned{color:gray}tr:nth-child(even) td{background:#F1F1F1}
                tr:nth-child(odd) td{background:#FEFEFE}tr td:hover{background:#666;color:#FFF}
                </style>
            </head>
            <body>
            """

            cursor = conn.execute("SELECT production, from_status, to_status, time from ProductionStatusAgentCache;")

            # Check if the results are non-empty
            if cursor.rowcount == 0:
                return S_OK()

            html_elements = ""
            for production, from_status, to_status, time in cursor:
                html_elements += (
                    "<tr>"
                    + "<td>"
                    + production
                    + "</td>"
                    + "<td class='"
                    + from_status
                    + "'>"
                    + from_status
                    + "</td>"
                    + "<td class='"
                    + to_status
                    + "'>"
                    + to_status
                    + "</td>"
                    + "<td>"
                    + time
                    + "</td>"
                    + "</tr>"
                )
            if html_elements:
                aggregated_body += """\
        <p class="setup">Transformations updated</p>
        <table>
            <tr>
                <th>Production</th>
                <th>From</th>
                <th>To</th>
                <th>Time</th>
            </tr>
            {html_elements}
        </table>
            """.format(
                    html_elements=html_elements
                )

            cursor = conn.execute("SELECT prod_requests, time from ProductionStatusAgentReqCache;")

            # Check if the results are non-empty
            if cursor.rowcount == 0:
                return S_OK()

            html_elements = ""
            for prod_requests, time in cursor:
                html_elements += "<tr>" + "<td>" + prod_requests + "</td>" + "<td>" + time + "</td>" + "</tr>"

            if html_elements:
                aggregated_body += """\
            <br />
            <p class="setup">Production Requests updated to Done status</p>
            <table>
            <tr>
                <th>Production Requests</th>
                <th>Time</th>
            </tr>
            {html_elements}
            </table>
        </body>
        </html>
        """.format(
                    html_elements=html_elements
                )

            aggregated_body += """\
        </body>
        </html>
        """

            res = self.diracAdmin.sendMail(
                "vladimir.romanovsky@cern.ch",
                "Transformation Status Updates",
                aggregated_body,
                "vladimir.romanovsky@cern.ch",
                html=True,
            )

            if not res["OK"]:
                self.log.error(f"Can't send email: {res['Message']}")
                return S_OK()

        # Delete anyway
        conn.execute("DELETE FROM ProductionStatusAgentCache;")
        conn.commit()
        conn.execute("VACUUM;")

        conn.execute("DELETE FROM ProductionStatusAgentReqCache;")
        conn.commit()
        conn.execute("VACUUM;")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for: LHCbDIRAC.AccountingSystem.private.Plotters.JobStepPlotter.

JobStepPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter

We are assuming there is a solid test of __bases__, we are not testing them
here and assuming they work fine.

IMPORTANT: the test MUST be pylint compliant !
"""
from unittest import mock
import unittest


class JobStepPlotterTestCase(unittest.TestCase):
    """JobStepPlotterTestCase."""

    moduleTested = None
    classsTested = None

    def mockModuleTested(self, moduleTested):
        """Used to not redo the mocking done on the parent class ( if any )"""

        # Tries to get the mocks of the parent TestCases ( if any )
        for baseClass in self.__class__.__bases__:
            try:
                # pylint: disable=no-member
                moduleTested = baseClass.mockModuleTested(moduleTested)
            except TypeError:
                continue

        # And then makes its own mock
        class MockJobStep:
            # pylint: disable=missing-docstring,no-init
            definitionKeyFields = (
                "JobGroup",
                "RunNumber",
                "EventType",
                "ProcessingType",
                "ProcessingStep",
                "Site",
                "FinalStepState",
            )

        moduleTested.JobStep = mock.Mock(return_value=MockJobStep())

        return moduleTested

    def setUp(self):
        """Setup the test case."""

        import LHCbDIRAC.AccountingSystem.private.Plotters.JobStepPlotter as moduleTested

        self.moduleTested = self.mockModuleTested(moduleTested)
        self.classsTested = self.moduleTested.JobStepPlotter

    def tearDown(self):
        """Tear down the test case."""

        del self.moduleTested
        del self.classsTested


# ...............................................................................


class JobStepPlotterUnitTest(JobStepPlotterTestCase):
    """JobStepPlotterUnitTest.

    <constructor>
     - test_instantiate
    <class variables>
     - test_typeName
     - test_typeKeyFields
     - test_reportCPUEfficiencyName
     - test_reportCPUUsageName
     - test_reportPieCPUTimeName
     - test_reportCumulativeCPUTimeName
     - test_reportNormCPUTimeName
     - test_reportCumulativeNormCPUTimeName
     - test_reportPieNormCPUTimeName
     - test_reportInputDataName
     - test_reportCumulativeInputDataName
     - test_reportPieInputDataName
     - test_reportOutputDataName
     - test_reportCumulativeOutputDataName
     - test_reportPieOutputDataName
     - test_reportInputEventsName
     - test_reportCumulativeInputEventsName
     - test_reportPieInputEventsName
     - test_reportOutputEventsName
     - test_reportCumulativeOutputEventsName
     - test_reportPieOutputEventsName
     - test_reportInputEventsPerOutputEventsName
     - test_reportCPUTimePerOutputEventsName
     - test_reportCPUTimePerInputEventsName
    <methods>
    """

    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""
        obj = self.classsTested(None, None)
        self.assertEqual("JobStepPlotter", obj.__class__.__name__, msg="Expected JobStepPlotter object")

    def test_typeName(self):
        """test the class variable "_typeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._typeName, "JobStep", msg="Expected JobStep as value")

    def test_typeKeyFields(self):
        """test the class variable "_typeKeyFields"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._typeKeyFields,
            ["JobGroup", "RunNumber", "EventType", "ProcessingType", "ProcessingStep", "Site", "FinalStepState"],
            msg="Expected keys from MockJobStep",
        )

    def test_reportCPUEfficiencyName(self):
        """test the class variable "_reportCPUEfficiencyName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportCPUEfficiencyName, "CPU efficiency", msg="Expected CPU efficiency as value")

    def test_reportCPUUsageName(self):
        """test the class variable "_reportCPUUsageName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportCPUUsageName, "CPU time", msg="Expected CPU time as value")

    def test_reportPieCPUTimeName(self):
        """test the class variable "_reportPieCPUTimeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportPieCPUTimeName, "Pie plot of CPU time", msg="Expected Pie plot of CPU time as value"
        )

    def test_reportCumulativeCPUTimeName(self):
        """test the class variable "_reportCumulativeCPUTimeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCumulativeCPUTimeName, "Cumulative CPU time", msg="Expected Cumulative CPU time as value"
        )

    def test_reportNormCPUTimeName(self):
        """test the class variable "_reportNormCPUTimeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportNormCPUTimeName, "NormCPU time", msg="Expected NormCPU time as value")

    def test_reportCumulativeNormCPUTimeName(self):
        """test the class variable "_reportCumulativeNormCPUTimeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCumulativeNormCPUTimeName,
            "Cumulative normalized CPU time",
            msg="Expected Cumulative normalized CPU time as value",
        )

    def test_reportPieNormCPUTimeName(self):
        """test the class variable "_reportPieNormCPUTimeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportPieNormCPUTimeName, "Pie plot of NormCPU time", msg="Expected Pie plot of NormCPU time as value"
        )

    def test_reportInputDataName(self):
        """test the class variable "_reportInputDataName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportInputDataName, "Input Data", msg="Expected Input Data as value")

    def test_reportCumulativeInputDataName(self):
        """test the class variable "_reportCumulativeInputDataName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCumulativeInputDataName, "Cumulative Input Data", msg="Expected Cumulative Input Data as value"
        )

    def test_reportPieInputDataName(self):
        """test the class variable "_reportPieInputDataName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportPieInputDataName, "Pie plot of Input Data", msg="Expected Pie plot of Input Data as value"
        )

    def test_reportOutputDataName(self):
        """test the class variable "_reportOutputDataName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportOutputDataName, "Output Data", msg="Expected Output Data as value")

    def test_reportCumulativeOutputDataName(self):
        """test the class variable "_reportCumulativeOutputDataName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCumulativeOutputDataName, "Cumulative OutputData", msg="Expected Cumulative OutputData as value"
        )

    def test_reportPieOutputDataName(self):
        """test the class variable "_reportPieOutputDataName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportPieOutputDataName, "Pie plot of Output Data", msg="Expected Pie plot of Output Data as value"
        )

    def test_reportInputEventsName(self):
        """test the class variable "_reportInputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportInputEventsName, "Input Events", msg="Expected Input Events as value")

    def test_reportCumulativeInputEventsName(self):
        """test the class variable "_reportCumulativeInputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCumulativeInputEventsName,
            "Cumulative Input Events",
            msg="Expected Cumulative Input Events as value",
        )

    def test_reportPieInputEventsName(self):
        """test the class variable "_reportPieInputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportPieInputEventsName, "Pie plot of Input Events", msg="Expected Pie plot of Input Events as value"
        )

    def test_reportOutputEventsName(self):
        """test the class variable "_reportOutputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportOutputEventsName, "Output Events", msg="Expected Output Events as value")

    def test_reportCumulativeOutputEventsName(self):
        """test the class variable "_reportCumulativeOutputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCumulativeOutputEventsName,
            "Cumulative Output Events",
            msg="Expected Cumulative Output Events as value",
        )

    def test_reportPieOutputEventsName(self):
        """test the class variable "_reportPieOutputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportPieOutputEventsName,
            "Pie plot of Output Events",
            msg="Expected Pie plot of Output Events as value",
        )

    def test_reportInputEventsPerOutputEventsName(self):
        """test the class variable "_reportInputEventsPerOutputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportInputEventsPerOutputEventsName,
            "Input/Output Events",
            msg="Expected Input/Output Events as value",
        )

    def test_reportCPUTimePerOutputEventsName(self):
        """test the class variable "_reportCPUTimePerOutputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCPUTimePerOutputEventsName,
            "CPUTime/Output Events",
            msg="Expected CPUTime/Output Events as value",
        )

    def test_reportCPUTimePerInputEventsName(self):
        """test the class variable "_reportCPUTimePerInputEventsName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._reportCPUTimePerInputEventsName, "CPUTime/Input Events", msg="Expected CPUTime/Input Events as value"
        )

    # FIXME: add crashes !


################################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(JobStepPlotterTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(JobStepPlotterUnitTest))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client test."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest

from unittest.mock import MagicMock

from DIRAC import S_OK, gLogger

from LHCbDIRAC.TransformationSystem.Client.TaskManager import LHCbWorkflowTasks
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import PluginUtilities, groupByRun


def getSitesForSE(ses):
    if ses == "pippo":
        return {"OK": True, "Value": ["Site2", "Site3"]}
    else:
        return {"OK": True, "Value": ["Site3"]}


class ClientTestCase(unittest.TestCase):
    def setUp(self):
        tcMock = MagicMock()
        sc = MagicMock()
        jmc = MagicMock()

        self.l_wft = LHCbWorkflowTasks(tcMock, submissionClient=sc, jobMonitoringClient=jmc)
        self.tc = TransformationClient()
        self.tc.dataProcessingTypes = ["MCSimulation", "DataReconstruction"]

        self.tsMock = MagicMock()

        self.fcMock = MagicMock()
        self.fcMock.getFileSize.return_value = S_OK({"Failed": [], "Successful": cachedLFNSize})

        gLogger.setLevel("DEBUG")

        self.maxDiff = None

    def tearDown(self):
        pass


class TaskManagerSuccess(ClientTestCase):
    pass


class TransformationClientSuccess(ClientTestCase):
    def test__applyTransformationFilesStateMachine(self):
        tsFiles = {}
        dictOfNewLFNsStatus = {}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {})

        tsFiles = {}
        dictOfNewLFNsStatus = {"foo": ["status", 2, 1234]}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {})

        tsFiles = {"foo": ["status", 2, 1234]}
        dictOfNewLFNsStatus = {"foo": "status"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {})

        tsFiles = {"foo": ["status", 2, 1234]}
        dictOfNewLFNsStatus = {"foo": "statusA"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"foo": "statusA"})

        tsFiles = {"foo": ["status", 2, 1234], "bar": ["status", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "status"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {})

        tsFiles = {"foo": ["status", 2, 1234], "bar": ["status", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "statusA"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"foo": "statusA"})

        tsFiles = {"foo": ["status", 2, 1234], "bar": ["status", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "A", "bar": "B"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"foo": "A", "bar": "B"})

        tsFiles = {"foo": ["status", 2, 1234]}
        dictOfNewLFNsStatus = {"foo": "A", "bar": "B"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"foo": "A"})

        tsFiles = {"foo": ["Assigned", 2, 1234]}
        dictOfNewLFNsStatus = {"foo": "A", "bar": "B"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"foo": "A"})

        tsFiles = {"foo": ["Assigned", 2, 1234], "bar": ["Assigned", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "Assigned", "bar": "Processed"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"bar": "Processed"})

        tsFiles = {"foo": ["Processed", 2, 1234], "bar": ["Unused", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "Assigned", "bar": "Processed"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"bar": "Processed"})

        tsFiles = {"foo": ["Processed", 2, 1234], "bar": ["Unused", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "Assigned", "bar": "Processed"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, True)
        self.assertEqual(res, {"foo": "Assigned", "bar": "Processed"})

        tsFiles = {"foo": ["Processed", 2, 1234], "bar": ["Unused", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "Assigned", "bar": "Processed"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, True)
        self.assertEqual(res, {"foo": "Assigned", "bar": "Processed"})

        tsFiles = {"foo": ["MaxReset", 12, 1234], "bar": ["Processed", 22, 5678]}
        dictOfNewLFNsStatus = {"foo": "Unused", "bar": "Unused"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {})

        tsFiles = {"foo": ["MaxReset", 12, 1234], "bar": ["Processed", 22, 5678]}
        dictOfNewLFNsStatus = {"foo": "Unused", "bar": "Unused"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, True)
        self.assertEqual(res, {"foo": "Unused", "bar": "Unused"})

        tsFiles = {"foo": ["Assigned", 20, 1234], "bar": ["Processed", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "Unused", "bar": "Unused"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, False)
        self.assertEqual(res, {"foo": "MaxReset"})

        tsFiles = {"foo": ["Assigned", 20, 1234], "bar": ["Processed", 2, 5678]}
        dictOfNewLFNsStatus = {"foo": "Unused", "bar": "Unused"}
        res = self.tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, True)
        self.assertEqual(res, {"foo": "Unused", "bar": "Unused"})

    def test__applyTransformationStatusStateMachine(self):
        transIDAsDict = {123: ["Active", "MCSimulation"]}
        dictOfProposedstatus = {123: "Stopped"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "Stopped")

        transIDAsDict = {123: ["New", "MCSimulation"]}
        dictOfProposedstatus = {123: "Active"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "Active")

        transIDAsDict = {123: ["New", "MCSimulation"]}
        dictOfProposedstatus = {123: "New"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "New")

        transIDAsDict = {123: ["New", "MCSimulation"]}
        dictOfProposedstatus = {123: "Stopped"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "New")

        transIDAsDict = {123: ["New", "MCSimulation"]}
        dictOfProposedstatus = {123: "Stopped"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, True)
        self.assertEqual(res, "Stopped")

        transIDAsDict = {123: ["New", "MCSimulation"]}
        dictOfProposedstatus = {123: "Idle"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "New")

        transIDAsDict = {123: ["Active", "MCSimulation"]}
        dictOfProposedstatus = {123: "Idle"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "Idle")

        transIDAsDict = {123: ["Active", "MCSimulation"]}
        dictOfProposedstatus = {123: "Completed"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "Flush")

        transIDAsDict = {123: ["Active", "MCSimulation"]}
        dictOfProposedstatus = {123: "Complete"}
        res = self.tc._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, False)
        self.assertEqual(res, "Active")


# Test data for plugins
data = {
    "/this/is/at_1": ["SE1"],
    "/this/is/at_2": ["SE2"],
    "/this/is/at_12": ["SE1", "SE2"],
    "/this/is/also/at_12": ["SE1", "SE2"],
    "/this/is/at_123": ["SE1", "SE2", "SE3"],
    "/this/is/at_23": ["SE2", "SE3"],
    "/this/is/at_4": ["SE4"],
}

cachedLFNSize = {
    "/this/is/at_1": 1,
    "/this/is/at_2": 2,
    "/this/is/at_12": 12,
    "/this/is/also/at_12": 12,
    "/this/is/at_123": 123,
    "/this/is/at_23": 23,
    "/this/is/at_4": 4,
}

runFiles = [
    {"LFN": "/this/is/at_1", "RunNumber": 1},
    {"LFN": "/this/is/at_2", "RunNumber": 2},
    {"LFN": "/this/is/at_12", "RunNumber": 12},
    {"LFN": "/this/is/also/at_12", "RunNumber": 12},
    {"LFN": "/this/is/at_123", "RunNumber": 123},
    {"LFN": "/this/is/at_23", "RunNumber": 23},
    {"LFN": "/this/is/at_4", "RunNumber": 4},
]

cachedLFNAncestors = {1: {"": 1}}


class PluginsUtilitiesSuccess(ClientTestCase):
    def test_groupByRun(self):
        # no files, nothing happens
        resDict = groupByRun([])
        self.assertEqual(resDict, {})

        # some files
        resDict = groupByRun(list(runFiles))
        resExpected = {
            1: ["/this/is/at_1"],
            2: ["/this/is/at_2"],
            4: ["/this/is/at_4"],
            12: ["/this/is/at_12", "/this/is/also/at_12"],
            23: ["/this/is/at_23"],
            123: ["/this/is/at_123"],
        }
        self.assertEqual(resDict, resExpected)

    def test_groupByRunAndParam(self):
        # no files, nothing happens
        pu = PluginUtilities(fc=self.fcMock, dataManager=MagicMock(), rmClient=MagicMock())
        pu.transFiles = []
        resDict = pu.getFilesGroupedByRunAndParam({})
        self.assertEqual(resDict, {})

        # some files, no params (it seems it's never called with params...?
        pu = PluginUtilities(fc=self.fcMock, dataManager=MagicMock(), rmClient=MagicMock())
        pu.transFiles = runFiles
        resDict = pu.getFilesGroupedByRunAndParam(data)
        resExpected = {
            1: {None: ["/this/is/at_1"]},
            2: {None: ["/this/is/at_2"]},
            4: {None: ["/this/is/at_4"]},
            12: {None: ["/this/is/at_12", "/this/is/also/at_12"]},
            23: {None: ["/this/is/at_23"]},
            123: {None: ["/this/is/at_123"]},
        }
        self.assertEqual(resDict, resExpected)

    def test_getRAWAncestorsForRun(self):
        # no files, nothing happens
        #     tsMock = MagicMock()
        #     tsMock.getTransformationFiles.return_value=S_OK( [] )
        #     pu = PluginUtilities(transClient=tsMock, fc=self.fcMock, dataManager=MagicMock(), rmClient=MagicMock() )
        #     res = pu.getRAWAncestorsForRun( 1 )
        #     self.assertEqual( res, 0 )

        # some files
        #     tsMock = MagicMock()
        #     tsMock.getTransformationFiles.return_value = S_OK( [{'LFN':'this/is/at_1', 'Status':'Unused'},
        #                                                         {'LFN':'this/is/not_here', 'Status':'MissingInFC'}] )
        #     bkMock = MagicMock()
        #     bkMock.getFileAncestors.return_value =
        #     pu = PluginUtilities( transClient=tsMock, fc=self.fcMock, dataManager=MagicMock(), rmClient=MagicMock() )
        #     res = pu.getRAWAncestorsForRun( 1 )
        #     self.assertEqual( res, 0 )
        pass


#############################################################################
#############################################################################
if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ClientTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TaskManagerSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TransformationClientSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PluginsUtilitiesSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import DIRAC
from DIRAC.Core.Base.Script import Script


def getBoolean(value):
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    else:
        print("ERROR: expected boolean")
        DIRAC.exit(2)


def usage():
    """usage.

    Prints script usage
    """
    print(f"Usage: {Script.scriptName} <ProductionID> <Options> [Try -h,--help for more information]")
    DIRAC.exit(2)


@Script()
def main():
    # Default values
    status = None
    outFile = None
    summary = True
    printVerbose = False

    Script.registerSwitch("", "Status=", "ProductionDB file status to select")
    Script.registerSwitch("", "OutputFile=", "Output file to store file records")
    Script.registerSwitch("", "Summary=", "Print a summary of the files (True/False) default is True")
    Script.registerSwitch("", "PrintOutput=", "Print all file records (extremely verbose) default is False")
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "status":
            status = switch[1]
        elif switch[0].lower() == "outputfile":
            outFile = switch[1]
        elif switch[0].lower() == "summary":
            summary = getBoolean(switch[1])
        elif switch[0].lower() == "printoutput":
            printVerbose = getBoolean(switch[1])

    args = Script.getPositionalArgs()

    if len(args) != 1:
        usage()

    exitCode = 0

    diracProd = DiracProduction()

    productionID = args[0]
    try:
        productionID = int(productionID)
    except Exception as x:
        print(f"Production ID must be an integer, not {productionID}:\n{x}")
        DIRAC.exit(2)

    result = diracProd.productionFileSummary(
        productionID,
        selectStatus=status,
        outputFile=outFile,
        printSummary=summary,
        printOutput=printVerbose,
    )
    if not result["OK"]:
        print(f"ERROR {result['Message']}")
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Returns list of TCKs for a run range, by default only if there is a FULL stream."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import Script

    Script.registerSwitch("", "Runs=", "Run range or list")
    Script.registerSwitch("", "ByRange", "List by range rather than by item value")
    Script.registerSwitch("", "Force", "Include runs even if no FULL stream is present")
    Script.registerSwitch("", "DQFlag=", "Specify the DQ flag (default: all)")
    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile] ... "]))

    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeRunInfo

    executeRunInfo("Tck")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It returns the input and output files of a given list of DIRAC Jobids."""
import os
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, printDMResult

    bkScript = DMScript()
    bkScript.registerJobsSwitches()
    Script.registerSwitch("", "InputFiles", "  Only input files")
    Script.registerSwitch("", "OutputFiles", "  Only output files")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... [DIRACJobid|File]",
                "Arguments:",
                "  DIRACJobid:      DIRAC Jobids",
                "  File:     Name of the file with contains a list of DIRACJobids",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    inputFiles = False
    outputFiles = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "InputFiles":
            inputFiles = True
        if switch[0] == "OutputFiles":
            outputFiles = True
    if not inputFiles and not outputFiles:
        inputFiles = True
        outputFiles = True
    jobidList = []
    for jobid in args:
        if os.path.exists(jobid):
            bkScript.setJobidsFromFile(jobid)
        else:
            jobidList += jobid.split(",")
    jobidList += bkScript.getOption("JobIDs", [])
    if not jobidList:
        print("No jobID provided!")
        Script.showHelp(exitCode=1)

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    retVal = BookkeepingClient().getJobInputOutputFiles(jobidList)
    if retVal["OK"]:
        success = retVal["Value"]["Successful"]
        for job in success:
            # Remove from input the files that are also output!
            # This happens because the output of step 1 can be the input of step 2...
            # only worth if input files are requested though
            if inputFiles:
                success[job]["InputFiles"] = sorted(set(success[job]["InputFiles"]) - set(success[job]["OutputFiles"]))

            if not inputFiles or not outputFiles:
                success[job].pop("InputFiles" if not inputFiles else "OutputFiles")

    printDMResult(retVal, empty="File does not exists in the Bookkeeping")


if __name__ == "__main__":
    main()
